@@ -1,58 +1,35 @@
 ;;;; slib.scm --- definitions needed to get SLIB to work with Guile
 ;;;;
-;;;;	Copyright (C) 1997, 1998, 2000, 2001, 2002, 2003, 2004 Free Software Foundation, Inc.
+;;;;	Copyright (C) 1997, 1998, 2000, 2001, 2002, 2003, 2004, 2006 Free Software Foundation, Inc.
 ;;;;
-;;;; This file is part of GUILE.
+;;;; This library is free software; you can redistribute it and/or
+;;;; modify it under the terms of the GNU Lesser General Public
+;;;; License as published by the Free Software Foundation; either
+;;;; version 2.1 of the License, or (at your option) any later version.
 ;;;; 
-;;;; GUILE is free software; you can redistribute it and/or modify it
-;;;; under the terms of the GNU General Public License as published by
-;;;; the Free Software Foundation; either version 2, or (at your
-;;;; option) any later version.
-;;;; 
-;;;; GUILE is distributed in the hope that it will be useful, but
-;;;; WITHOUT ANY WARRANTY; without even the implied warranty of
+;;;; This library is distributed in the hope that it will be useful,
+;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-;;;; General Public License for more details.
+;;;; Lesser General Public License for more details.
 ;;;; 
-;;;; You should have received a copy of the GNU General Public License
-;;;; along with GUILE; see the file COPYING.  If not, write to the
-;;;; Free Software Foundation, Inc., 59 Temple Place, Suite 330,
-;;;; Boston, MA 02111-1307 USA
-;;;;
-;;;; As a special exception, the Free Software Foundation gives permission
-;;;; for additional uses of the text contained in its release of GUILE.
-;;;;
-;;;; The exception is that, if you link the GUILE library with other files
-;;;; to produce an executable, this does not by itself cause the
-;;;; resulting executable to be covered by the GNU General Public License.
-;;;; Your use of that executable is in no way restricted on account of
-;;;; linking the GUILE library code into it.
-;;;;
-;;;; This exception does not however invalidate any other reasons why
-;;;; the executable file might be covered by the GNU General Public License.
-;;;;
-;;;; This exception applies only to the code released by the
-;;;; Free Software Foundation under the name GUILE.  If you copy
-;;;; code from other Free Software Foundation releases into a copy of
-;;;; GUILE, as the General Public License permits, the exception does
-;;;; not apply to the code that you add in this way.  To avoid misleading
-;;;; anyone as to the status of such modified files, you must delete
-;;;; this exception notice from them.
-;;;;
-;;;; If you write modifications of your own for GUILE, it is your choice
-;;;; whether to permit this exception to apply to your modifications.
-;;;; If you do not wish that, delete this exception notice.
+;;;; You should have received a copy of the GNU Lesser General Public
+;;;; License along with this library; if not, write to the Free Software
+;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 ;;;;
 (define-module (ice-9 slib)
-  :export (slib:load
-	   implementation-vicinity
-	   library-vicinity
-	   home-vicinity
-	   scheme-implementation-type
-	   scheme-implementation-version
+  :export (slib:load slib:load-source defmacro:load
+	   implementation-vicinity library-vicinity home-vicinity
+	   scheme-implementation-type scheme-implementation-version
+	   output-port-width output-port-height array-indexes
 	   make-random-state
-	   <? <=? =? >? >=?
-	   require)
+	   -1+ <? <=? =? >? >=?
+	   require slib:error slib:exit slib:warn slib:eval
+	   defmacro:eval logical:logand logical:logior logical:logxor
+	   logical:lognot logical:ash logical:logcount logical:integer-length
+	   logical:bit-extract logical:integer-expt logical:ipow-by-squaring
+	   slib:eval-load slib:tab slib:form-feed difftime offset-time
+	   software-type)
+  :replace (delete-file open-file provide provided? system)
   :no-backtrace)
 
 
@@ -71,10 +48,10 @@
 
 
 
-(define-public slib:exit quit)
-(define-public slib:error error)
-(define-public slib:warn warn)
-(define-public slib:eval (lambda (x) (eval x slib-module)))
+(define slib:exit quit)
+(define slib:error error)
+(define slib:warn warn)
+(define slib:eval (lambda (x) (eval x slib-module)))
 (define defmacro:eval (lambda (x) (eval x (interaction-environment))))
 (define logical:logand logand)
 (define logical:logior logior)
@@ -85,74 +62,101 @@
 (define logical:integer-length integer-length)
 (define logical:bit-extract bit-extract)
 (define logical:integer-expt integer-expt)
-(define logical:ipow-by-squaring ipow-by-squaring)
-(define-public slib:eval-load eval-load)
-(define-public slib:tab #\tab)
-(define-public slib:form-feed #\page)
+(define slib:eval-load eval-load)
+(define slib:tab #\tab)
+(define slib:form-feed #\page)
 
 (define slib-module (current-module))
 
 (define (defined? symbol)
   (module-defined? slib-module symbol))
 
-(define slib:features
-  (append '(source
-	    eval
-	    abort
-	    alist
-	    defmacro
-	    delay
-	    dynamic-wind
-	    full-continuation
-	    hash
-	    hash-table
-	    line-i/o
-	    logical
-	    multiarg/and-
-	    multiarg-apply
-	    promise
-	    rev2-procedures
-	    rev4-optional-procedures
-	    string-port
-	    with-file)
+;;; *FEATURES* should be set to a list of symbols describing features
+;;; of this implementation.  Suggestions for features are:
+(set! *features*
+  (append
+      '(
+	source				;can load scheme source files
+					;(slib:load-source "filename")
+;	compiled			;can load compiled files
+					;(slib:load-compiled "filename")
 
-	  (if (defined? 'getenv)
-	      '(getenv)
-	      '())
+		       ;; Scheme report features
 
-	  (if (defined? 'current-time)
-	      '(current-time)
-	      '())
+;	rev5-report			;conforms to
+	eval				;R5RS two-argument eval
+;	values				;R5RS multiple values
+	dynamic-wind			;R5RS dynamic-wind
+;	macro				;R5RS high level macros
+	delay				;has DELAY and FORCE
+	multiarg-apply			;APPLY can take more than 2 args.
+;	rationalize
+	rev4-optional-procedures	;LIST-TAIL, STRING->LIST,
+					;LIST->STRING, STRING-COPY,
+					;STRING-FILL!, LIST->VECTOR,
+					;VECTOR->LIST, and VECTOR-FILL!
 
-	  (if (defined? 'system)
-	      '(system)
-	      '())
+;	rev4-report			;conforms to
 
-	  (if (defined? 'char-ready?)
-	      '(char-ready?)
-	      '())
+;	ieee-p1178			;conforms to
 
-	  (if (and (string->number "0.0") (inexact? (string->number "0.0")))
-	      '(inexact)
-	      '())
+;	rev3-report			;conforms to
 
-	  (if (rational? (string->number "1/19"))
-	      '(rational)
-	      '())
+	rev2-procedures			;SUBSTRING-MOVE-LEFT!,
+					;SUBSTRING-MOVE-RIGHT!,
+					;SUBSTRING-FILL!,
+					;STRING-NULL?, APPEND!, 1+,
+					;-1+, <?, <=?, =?, >?, >=?
+;	object-hash			;has OBJECT-HASH
 
-	  (if (real? (string->number "0.0"))
-	      '(real)
-	      ())
+	multiarg/and-			;/ and - can take more than 2 args.
+	with-file			;has WITH-INPUT-FROM-FILE and
+					;WITH-OUTPUT-FROM-FILE
+;	transcript			;TRANSCRIPT-ON and TRANSCRIPT-OFF
+;	ieee-floating-point		;conforms to IEEE Standard 754-1985
+					;IEEE Standard for Binary
+					;Floating-Point Arithmetic.
+	full-continuation		;can return multiple times
 
-	  (if (complex? (string->number "1+i"))
-	      '(complex)
-	      '())
+			;; Other common features
 
-	  (let ((n (string->number "9999999999999999999999999999999")))
-	    (if (and n (exact? n))
-		'(bignum)
-		'()))))
+;	srfi				;srfi-0, COND-EXPAND finds all srfi-*
+;	sicp				;runs code from Structure and
+					;Interpretation of Computer
+					;Programs by Abelson and Sussman.
+	defmacro			;has Common Lisp DEFMACRO
+;	record				;has user defined data structures
+	string-port			;has CALL-WITH-INPUT-STRING and
+					;CALL-WITH-OUTPUT-STRING
+;	sort
+;	pretty-print
+;	object->string
+;	format				;Common-lisp output formatting
+;	trace				;has macros: TRACE and UNTRACE
+;	compiler			;has (COMPILER)
+;	ed				;(ED) is editor
 
+	;; core definitions compatible, plus `make-random-state' below
+	random
+	)
+
+	(if (defined? 'getenv)
+	    '(getenv)
+	    '())
+
+	(if (defined? 'current-time)
+	    '(current-time)
+	    '())
+
+	(if (defined? 'system)
+	    '(system)
+	    '())
+
+	(if (defined? 'char-ready?)
+	    '(char-ready?)
+	    '())
+
+	*features*))
 
 ;; The array module specified by slib 3a1 is not the same as what guile
 ;; provides, so we must remove `array' from the features list.
@@ -183,6 +187,21 @@
 (set! *features* (delq 'array          *features*))
 (set! *features* (delq 'array-for-each *features*))
 
+;; The random module in slib 3a1 provides a `random:chunk' which is used by
+;; the random-inexact module.  Guile doesn't provide random:chunk so we must
+;; remove 'random from `*features*' to use the slib code.
+;;
+;; ENHANCE-ME: Maybe Guile could provide a `random:chunk', the rest of the
+;; random module is already the same as Guile.
+;;
+;; FIXME: As per the array bits above, *features* is shared by slib and the
+;; guile core, so removing 'random has the unhappy effect of making it look
+;; like this isn't in the core.  Let's assume random numbers have been
+;; present unconditionally long enough that no guile-specific code will
+;; bother to test.
+;;
+(set! *features* (delq 'random *features*))
+
 
 ;;; FIXME: Because uers want require to search the path, this uses
 ;;; load-from-path, which probably isn't a hot idea.  slib
@@ -211,7 +230,7 @@
 		       (lambda args args)))
 	   (apply throw errinfo))))))
 
-(define-public slib:load-source slib:load)
+(define slib:load-source slib:load)
 (define defmacro:load slib:load)
 
 (define slib-parent-dir
@@ -228,16 +247,22 @@
   (let ((home-path (getenv "HOME")))
     (lambda () home-path)))
 (define (scheme-implementation-type) 'guile)
-(define (scheme-implementation-version) "")
+(define scheme-implementation-version version)
+;;; (scheme-implementation-home-page) should return a (string) URI
+;;; (Uniform Resource Identifier) for this scheme implementation's home
+;;; page; or false if there isn't one.
+(define (scheme-implementation-home-page)
+  "http://www.gnu.org/software/guile/guile.html")
 
 ;; legacy from r3rs, but slib says all implementations provide these
 ;; ("Legacy" section of the "Miscellany" node in the manual)
 (define-public t   #t)
 (define-public nil #f)
 
-(define-public (output-port-width . arg) 80)
-(define-public (output-port-height . arg) 24)
-(define (identity x) x)
+;; ENHANCE-ME: Could call ioctl TIOCGWINSZ to get the size of a tty (see
+;; "man 4 tty_ioctl" on a GNU/Linux system), on systems with that.
+(define (output-port-width . arg) 80)
+(define (output-port-height . arg) 24)
 
 ;; slib 3a1 and up, straight from Template.scm
 (define-public (call-with-open-ports . ports)
@@ -257,20 +282,16 @@
 ;; sure if that's intentional, but in any case this extension continues to
 ;; accept strings to make that work.
 ;;
-(define-public open-file
-  (let ((guile-core-open-file open-file))
-    (lambda (filename modes)
-      (if (symbol? modes)
-	  (set! modes (symbol->string modes)))
-      (guile-core-open-file filename modes))))
+(define-public (open-file filename modes)
+  (if (symbol? modes)
+      (set! modes (symbol->string modes)))
+  ((@ (guile) open-file) filename modes))
 
 ;; returning #t/#f instead of throwing an error for failure
-(define-public delete-file
-  (let ((guile-core-delete-file delete-file))
-    (lambda (filename)
-      (catch 'system-error
-	(lambda () (guile-core-delete-file filename) #t)
-	(lambda args #f)))))
+(define-public (delete-file filename)
+  (catch 'system-error
+    (lambda () ((@ (guile) delete-file) filename) #t)
+    (lambda args #f)))
 
 ;; Nothing special to do for this, so straight from Template.scm.  Maybe
 ;; "sensible-browser" for a debian system would be worth trying too (and
@@ -282,6 +303,12 @@
       (try "netscape '" "'&")
       (try "netscape '" "'")))
 
+;;; {array-for-each}
+(define (array-indexes ra)
+  (let ((ra0 (apply make-array '() (array-shape ra))))
+    (array-index-map! ra0 list)
+    ra0))
+
 ;;; {Random numbers}
 ;;;
 (define (make-random-state . args)
@@ -296,6 +323,7 @@
 ;;; {rev2-procedures}
 ;;;
 
+(define -1+ 1-)
 (define <?  <)
 (define <=? <=)
 (define =?  =)
@@ -304,7 +332,6 @@
 
 ;;; {system}
 ;;;
-
 ;; If the program run is killed by a signal, the shell normally gives an
 ;; exit code of 128+signum.  If the shell itself is killed by a signal then
 ;; we do the same 128+signum here.
@@ -312,30 +339,27 @@
 ;; "stop-sig" shouldn't arise here, since system shouldn't be calling
 ;; waitpid with WUNTRACED, but allow for it anyway, just in case.
 ;;
-(if (defined? 'system)
+(if (memq 'system *features*)
     (define-public system
-      (let ((guile-core-system system))
-	(lambda (str)
-	  (let ((st (guile-core-system str)))
-	    (or (status:exit-val st)
-		(+ 128 (or (status:term-sig st)
-			   (status:stop-sig st)))))))))
+      (lambda (str)
+	(let ((st ((@ (guile) system) str)))
+	  (or (status:exit-val st)
+	      (+ 128 (or (status:term-sig st)
+			 (status:stop-sig st))))))))
 
 ;;; {Time}
 ;;;
 
-(define-public difftime -)
-(define-public offset-time +)
+(define difftime -)
+(define offset-time +)
 
 
-(define %system-define define)
-
 (define define
   (procedure->memoizing-macro
    (lambda (exp env)
      (if (= (length env) 1)
 	 `(define-public ,@(cdr exp))
-	 `(%system-define ,@(cdr exp))))))
+	 `(define-private ,@(cdr exp))))))
 
 ;;; Hack to make syncase macros work in the slib module
 (if (nested-ref the-root-module '(app modules ice-9 syncase))
