@@ -1,9 +1,9 @@
 /*
  * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
- * Copyright (c) 1991-1995, 2003 by Xerox Corporation.  All rights reserved.
+ * Copyright (c) 1991-1995 by Xerox Corporation.  All rights reserved.
  * Copyright (c) 1996-1999 by Silicon Graphics.  All rights reserved.
  * Copyright (c) 1999 by Hewlett-Packard Company.  All rights reserved.
- * Copyright (c) 2000, 2001, 2003 Free Software Foundation
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2006 Free Software Foundation
  *
  * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
  * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
@@ -21,6 +21,10 @@
  * and modified for Guile by Marius Vollmer.
  */
 
+#if HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
 #include <ctype.h>
 #include "libguile/gc.h"
 #include "libguile/scmconfig.h"
@@ -37,7 +41,6 @@ scm_get_stack_base ()
 
 #else
 
-
 #define ABORT(msg) abort ()
 
 typedef char * ptr_t;	/* A generic pointer to which we can add	*/
@@ -112,12 +115,14 @@ typedef int GC_bool;
 #    define NETBSD
 #    define mach_type_known
 # endif
-# if defined(__NetBSD__) && defined(m68k)
+/* in netbsd 2.0 only __m68k__ is defined, not m68k */
+# if defined(__NetBSD__) && (defined(m68k) || defined(__m68k__))
 #    define M68K
 #    define NETBSD
 #    define mach_type_known
 # endif
-# if defined(__NetBSD__) && defined(arm32)
+/* in netbsd 2.0 only __arm__ is defined, not arm32 */
+# if defined(__NetBSD__) && (defined(arm32) || defined(__arm__))
 #    define ARM32
 #    define NETBSD
 #    define mach_type_known
@@ -235,7 +240,7 @@ typedef int GC_bool;
 #    define SPARC
 #    define mach_type_known
 # endif
-# if defined(LINUX) && (defined(arm) || defined(__arm__))
+# if defined(LINUX) && (defined(arm) || defined (__arm__))
 #    define ARM32
 #    define mach_type_known
 # endif
@@ -460,7 +465,7 @@ scm_get_stack_base ()
  * highest address in the stack.
  * Under PCR or OS/2, we have other ways of finding thread stacks.
  * For each machine, the following should:
- * 1) define STACK_GROWS_UP if the stack grows toward higher addresses, and
+ * 1) define SCM_STACK_GROWS_UP if the stack grows toward higher addresses, and
  * 2) define exactly one of
  *	STACKBOTTOM (should be defined to be an expression)
  *	HEURISTIC1
@@ -470,7 +475,7 @@ scm_get_stack_base ()
  * HEURISTIC1:  Take an address inside GC_init's frame, and round it up to
  *		the next multiple of STACK_GRAN.
  * HEURISTIC2:  Take an address inside GC_init's frame, increment it repeatedly
- *		in small steps (decrement if STACK_GROWS_UP), and read the value
+ *		in small steps (decrement if SCM_STACK_GROWS_UP), and read the value
  *		at each location.  Remember the value when the first
  *		Segmentation violation or Bus error is signalled.  Round that
  *		to the nearest plausible page boundary, and use that instead
@@ -1095,7 +1100,9 @@ scm_get_stack_base ()
 	extern char ** environ;
 #       define STACKBOTTOM ((ptr_t)environ)
 #   endif
-#   define STACK_GROWS_UP
+#   ifndef SCM_STACK_GROWS_UP /* don't fight with scmconfig.h */
+#     define SCM_STACK_GROWS_UP 1
+#   endif
 #   define DYNAMIC_LOADING
 #   ifndef HPUX_THREADS
 #     define MPROTECT_VDB
@@ -1281,9 +1288,11 @@ scm_get_stack_base ()
 #   define DATASTART ((ptr_t)(&__data_start != 0? &__data_start : &data_start))
 #endif
 
-# ifndef STACK_GROWS_UP
-#   define STACK_GROWS_DOWN
-# endif
+# if SCM_STACK_GROWS_UP
+#   define STACK_GROWS_DOWN 0
+# else
+#   define STACK_GROWS_DOWN 1
+#endif
 
 # ifndef CPP_WORDSZ
 #   define CPP_WORDSZ 32
@@ -1882,7 +1891,7 @@ void *scm_get_stack_base()
 	return(STACKBOTTOM);
 #   else
 #	ifdef HEURISTIC1
-#	   ifdef STACK_GROWS_DOWN
+#	   if STACK_GROWS_DOWN
 	     result = (ptr_t)((((word)(&dummy))
 	     		       + STACKBOTTOM_ALIGNMENT_M1)
 			      & ~STACKBOTTOM_ALIGNMENT_M1);
@@ -1895,7 +1904,7 @@ void *scm_get_stack_base()
 	   result = GC_linux_stack_base();
 #	endif
 #	ifdef HEURISTIC2
-#	    ifdef STACK_GROWS_DOWN
+#	    if STACK_GROWS_DOWN
 		result = GC_find_limit((ptr_t)(&dummy), TRUE);
 #           	ifdef HEURISTIC2_LIMIT
 		    if (result > HEURISTIC2_LIMIT
@@ -1914,7 +1923,7 @@ void *scm_get_stack_base()
 #	    endif
 
 #	endif /* HEURISTIC2 */
-#	ifdef STACK_GROWS_DOWN
+#	if STACK_GROWS_DOWN
 	    if (result == 0) result = (ptr_t)(signed_word)(-sizeof(ptr_t));
 #	endif
     	return(result);
